# Command-line interface for ColorVideoVDP. 

import os, sys
import os.path
import argparse
import logging
#from natsort import natsorted
import glob
import ffmpeg
import numpy as np
import torch
import imageio.v2 as imageio
import re
import inspect

from pycvvdp.vq_metric import vq_metric_dict

import pycvvdp

import shlex

#from pyfvvdp.fvvdp_display_model import fvvdp_display_photometry, fvvdp_display_geometry
# from pyfvvdp.visualize_diff_map import visualize_diff_map
import pycvvdp.utils as utils

from pycvvdp.ssim_metric import ssim_metric
from pycvvdp.dm_preview_metric import *
from pycvvdp.dump_channels import DumpChannels

def expand_wildcards(filestrs):
    if not isinstance(filestrs, list):
        return [ filestrs ]
    files = []
    for filestr in filestrs:
        if "*" in filestr:
            curlist = sorted( glob.glob(filestr) )
            files = files + curlist
        else:
            files.append(filestr)
    return files

# Save a numpy array as a video
def np2vid(np_srgb, vidfile, fps, verbose=False):

    N, H, W, C = np_srgb.shape
    if C == 1:
        np_srgb = np.concatenate([np_srgb]*3, -1)
    process = (
        ffmpeg
            .input('pipe:', format='rawvideo', pix_fmt='rgb24', s='{}x{}'.format(W, H), r=fps)
            #.output(vidfile, format='mp4', **{ "v:q": "10" } )
            .output(vidfile, format='mp4', **{ "c:v": "mpeg4", "qscale:v": "3" } )  # mpeg4 codec is always bundled with ffmpeg so it should work
            .overwrite_output()
            .global_args( '-hide_banner')
            .global_args( '-loglevel', 'info' if verbose else 'quiet')
            .run_async(pipe_stdin=True)
    )
    for fid in range(N):
        process.stdin.write(
                (np_srgb[fid,...] * 255.0)
                .astype(np.uint8)
                .tobytes()
        )
    process.stdin.close()
    process.wait()

# Save a numpy array as an image
def np2img(np_srgb, imgfile):

    N, H, W, C = np_srgb.shape
    if C == 1:
        np_srgb = np.concatenate([np_srgb]*3, -1)

    if N>1:
        sys.exit("Expecting an image, found video")

    imageio.imwrite( imgfile, (np.clip(np_srgb,0.0,1.0)[0,...]*255.0).astype(np.uint8) )

# -----------------------------------
# Command-line Arguments
# -----------------------------------
def parse_args(arg_list=None):

    available_metrics = [mm.replace('_', '-') for mm in vq_metric_dict.keys()] # Use dash in metric names

    parser = argparse.ArgumentParser(description="Evaluate ColorVideoVDP on a set of videos")
    parser.add_argument("-t", "--test", type=str, nargs='+', required = False, help="list of test images/videos")
    parser.add_argument("-r", "--ref", type=str, nargs='+', required = False, help="list of reference images/videos")
    parser.add_argument("--device", type=str,  default='cuda', help="select which PyTorch device to use. Pick from ['cpu', 'mps', 'cuda', 'cuda:0', 'cuda:1', ...]")
    parser.add_argument("--heatmap", type=str, default="none", help="type of difference map (none, raw, threshold, supra-threshold).")
    parser.add_argument("-g", "--distogram", type=float, default=-1, const=10, nargs='?', help="generate a distogram that visualizes the differences per-channel and per frame. The optional floating point parameter is the maximum JOD value to use in the visualization.")
    parser.add_argument("-x", "--features", action='store_true', default=False, help="generate JSON files with extracted features. Useful for retraining the metric.")
    parser.add_argument("-o", "--output-dir", type=str, default=None, help="in which directory heatmaps and feature files should be stored (the default is the current directory)")
    parser.add_argument("--result", type=str, default=None, help="write metric prediction results to a CSV file passed as an argument.")
    parser.add_argument("-c", "--config-paths", type=str, nargs='+', default=[], help="One or more paths to configuration files or directories. The main configurations files are `display_models.json`, `color_spaces.json` and `cvvdp_parameters.json`. The file name must start as the name of the original config file.")
    parser.add_argument("-d", "--display", type=str, default="standard_4k", help="display name, e.g. 'HTC Vive', or ? to print the list of models.")
    parser.add_argument("-n", "--nframes", type=int, default=-1, help="the number of video frames you want to compare")
    parser.add_argument("-f", "--full-screen-resize", choices=['bilinear', 'bicubic', 'nearest', 'area'], default=None, help="Both test and reference videos will be resized to match the full resolution of the display. Currently works only with videos.")
    parser.add_argument("-m", "--metric", choices=available_metrics, nargs='+', default=['cvvdp'], help='Select which metric(s) to run')
    parser.add_argument("--temp-padding", choices=['replicate', 'circular', 'pingpong'], default='replicate', help='How to pad the video in the time domain (for the temporal filters). "replicate" - repeat the first frame. "pingpong" - mirror the first frames. "circular" - take the last frames.')
    parser.add_argument("--pix-per-deg", type=float, default=None, help='Overwrite display geometry and use the provided pixels per degree value.')
    parser.add_argument("--fps", type=float, default=None, help='Frames per second. It will overwrite frame rate stores in the video file. Required when passing an array of image files.')
    parser.add_argument("--frames", type=str, default=None, help='Range of frames specified as first:step:last, first:last, or first: (Matlab notation). Currently works only with frames provided as images.')
    parser.add_argument("--gpu-mem", type=float, default=None, help='How much GPU memory can we use in GB. Use if CUDA reports out of mem errors, or you want to run multiple instances at the same time.')
    parser.add_argument("-q", "--quiet", action='store_true', default=False, help="Do not print any information but the final JOD value. Warning message will be still printed.")
    parser.add_argument("-v", "--verbose", action='store_true', default=False, help="Print out extra information.")
    parser.add_argument("--ffmpeg-cc", action='store_true', default=False, help="Use ffmpeg for upsampling and color conversion. Use custom pytorch code by default (faster and less memory).")    
    parser.add_argument("--temp-resample", action='store_true', default=False, help="Resample test and reference video to a common frame rate. Allows to compare videos of different frame rates.")
    parser.add_argument("-i", "--interactive", action='store_true', default=False, help="Run in an interactive mode, in which command line arguments are provided to the standard input, line by line. Saves on start-up time when running a large number of comparisons.")
    parser.add_argument("--dump-channels", nargs='+', choices=['temporal', 'lpyr', 'difference'], default=None, help="Output video/images with intermediate processing stages (for debugging and visualization).")
    if arg_list is not None:
        args = parser.parse_args(arg_list)
    else:
        args = parser.parse_args()
    return args

def run_on_args(args):
    if args.quiet:
        log_level = logging.ERROR
    else:        
        log_level = logging.DEBUG if args.verbose else logging.INFO
        
    logging.basicConfig(format='[%(levelname)s] %(message)s', level=log_level)

    args.metric = [mm.replace('-', '_') for mm in args.metric] # We need underscore for class names

    if args.verbose:
        import platform
        logging.debug( f'Platform: {platform.platform()}' )

    if args.display == "?":
        pycvvdp.vvdp_display_photometry.list_displays(args.config_paths)
        return

    if args.test is None or args.ref is None:
        logging.error( "Paths to both test and reference content needs to be specified.")
        return

    # Range of frames to process
    frame_range = None
    if not args.frames is None:
        ss = args.frames.split(':')
        if len(ss) == 3:
            sn = [0, 1, 10000] # default values
        else:
            sn = [0, 10000]
        for kk in range(len(ss)):
            if ss[kk].isnumeric():
                sn[kk] = int(ss[kk])
        if len(ss) == 3:
            frame_range = range(sn[0],(sn[2]+1),sn[1])
        elif len(ss) <= 2:
            frame_range = range(sn[0],(sn[1]+1))
        


    # Changed option to include MPS support for Macbooks
    # if args.gpu >= 0 and torch.cuda.is_available():
    #     device = torch.device('cuda:' + str(args.gpu))
    # else:
    #     device = torch.device('cpu')
    args.device = args.device.lower()
    if args.device.startswith('cuda') and torch.cuda.is_available():
        device = torch.device(args.device)
    elif args.device == 'mps':
        torch_version = list(map(int, re.search(r'\d+\.\d+\.\d+', torch.__version__).group(0).split('.')))
        assert torch_version[0] > 2 or (torch_version[0] == 2 and torch_version[1] > 0), f'Please use torch>=2.1.0 with MPS. Current version is {torch.__version__}.'
        logging.warn('MPS support is experimental. Please report any issues encountered.')
        assert sys.platform == 'darwin', 'Device "mps" is only valid on a Mac.'
        device = torch.device(args.device)
    else:
        if args.device != 'cpu':
            logging.warning(f'The requested device ({args.device}) is not found, reverting to CPU. This may result in slow execution.')
        device = torch.device('cpu')

    logging.info("Running on device: " + str(device))

    heatmap_types = ["raw", "threshold", "supra-threshold"]

    if args.heatmap == "none":
        args.heatmap = None

    if args.heatmap:
        if not args.heatmap in heatmap_types:
            logging.error( 'The recognized heatmap types are: "none", "raw", "threshold" and "supra-threshold"' )
            sys.exit()

        do_heatmap = True
    else:
        do_heatmap = False
        
    # Check for valid resizing methods
    #if args.gpu_decode:
        # Doc: https://pytorch.org/docs/stable/generated/torch.nn.functional.interpolate.html
        #valid_methods = ['nearest', 'bilinear', 'bicubic', 'area', 'nearest-exact']
    #else:
        # Doc: https://ffmpeg.org/ffmpeg-scaler.html
        #valid_methods = ['fast_bilinear', 'bilinear', 'bicubic', 'experimental', 'neighbor', 'area', 'bicublin', 'gauss', 'lanczos', 'spline']
    #if args.full_screen_size not in valid_methods:
    #    logging.error(f'The resizing method supplied is invalid. Please pick from {valid_methods}.')
    #    sys.exit()

    args.test = expand_wildcards(args.test)
    args.ref = expand_wildcards(args.ref)    

    N_test = len(args.test)
    N_ref = len(args.ref)

    if N_test==0:
        logging.error( "No test images/videos found." )
        sys.exit()

    if N_ref==0:
        logging.error( "No reference images/videos found." )
        sys.exit()

    if N_test != N_ref and N_test != 1 and N_ref != 1:
        logging.error( "Pass the same number of reference and test sources, or a single reference (to be used with all test sources), or a single test (to be used with all reference sources)." )
        sys.exit()

    metrics = []
    display_photometry = pycvvdp.vvdp_display_photometry.load(args.display, config_paths=args.config_paths)
    if args.pix_per_deg is None:
        display_geometry = pycvvdp.vvdp_display_geometry.load(args.display, config_paths=args.config_paths)
    else:
        display_geometry = pycvvdp.vvdp_display_geometry( [1024, 1024], ppd=args.pix_per_deg )

    out_dir = "." if args.output_dir is None else args.output_dir
    os.makedirs(out_dir, exist_ok=True)

    if args.dump_channels:
        dump_channels = DumpChannels( dump_temp_ch=("temporal" in args.dump_channels), dump_lpyr=("lpyr" in args.dump_channels), dump_diff=("difference" in args.dump_channels), output_dir=args.output_dir )
    else:
        dump_channels = None

    for mm in args.metric:
        if not mm in vq_metric_dict:
            raise RuntimeError( f"Unknown metric {mm}")
        metric_class = vq_metric_dict[mm]

        # The code below will figure out and pass only the parameters that a metric needs
        constructor_args = inspect.getfullargspec(metric_class.__init__)[0]
        all_bases = list(metric_class.__bases__)
        for bc in all_bases:
            constructor_args.extend( inspect.getfullargspec(bc.__init__)[0] )
            all_bases.extend( list(bc.__bases__) )

        met_args = {}
        if 'display_photometry' in constructor_args:
            met_args['display_photometry'] = display_photometry
        if 'display_geometry' in constructor_args:
            met_args['display_geometry'] = display_geometry
        if 'device' in constructor_args:
            met_args['device'] = device        
        if 'heatmap' in constructor_args:
            met_args['heatmap'] = args.heatmap        
        if 'temp_padding' in constructor_args:
            met_args['temp_padding'] = args.temp_padding        
        if 'config_paths' in constructor_args:
            met_args['config_paths'] = args.config_paths        
        if 'gpu_mem' in constructor_args:
            met_args['gpu_mem'] = args.gpu_mem        
        if 'dump_channels' in constructor_args:
            met_args['dump_channels'] = dump_channels        
        fv = metric_class(**met_args)
        fv.train(False)
        metrics.append( fv )

        info_str = metrics[-1].get_info_string()
        if not info_str is None:
            logging.info( 'When reporting metric results, please include the following information:' )
            logging.info( info_str )

    if not args.result is None:
        res_fh = open( args.result, "w" )
        res_fh.write( 'test, reference' )
        for mm in metrics:
            res_fh.write( ', ' + mm.short_name() )
        res_fh.write( '\n' )
    else:
        res_fh = None
        

    for kk in range( max(N_test, N_ref) ): # For each test and reference pair
        test_file = args.test[min(kk,N_test-1)]
        ref_file = args.ref[min(kk,N_ref-1)]
        if not res_fh is None:
            res_fh.write( f"{test_file}, {ref_file}" )
        logging.info(f"Predicting the quality of '{test_file}' compared to '{ref_file}'")
        for mm in metrics:
            preload = False if args.temp_padding == 'replicate' else True
            with torch.no_grad():

                if args.temp_resample:
                    vs = pycvvdp.video_source_temp_resample_file( test_file, ref_file, 
                                                display_photometry=display_photometry, 
                                                config_paths=args.config_paths,
                                                full_screen_resize=args.full_screen_resize, 
                                                resize_resolution=display_geometry.resolution, 
                                                frames=args.nframes,
                                                ffmpeg_cc=args.ffmpeg_cc,
                                                verbose=args.verbose )
                else:
                    vs = pycvvdp.video_source_file( test_file, ref_file, 
                                                display_photometry=display_photometry, 
                                                config_paths=args.config_paths,
                                                full_screen_resize=args.full_screen_resize, 
                                                resize_resolution=display_geometry.resolution, 
                                                frames=args.nframes,
                                                fps=args.fps,
                                                frame_range=frame_range,
                                                preload=preload,
                                                ffmpeg_cc=args.ffmpeg_cc,
                                                verbose=args.verbose )

                base, ext = os.path.splitext(os.path.basename(test_file))            
                base_fname = os.path.join(out_dir, base)
                mm.set_base_fname(base_fname)

                Q_pred, stats = mm.predict_video_source(vs)
                if args.quiet:
                    print( "{Q:0.4f}".format(Q=Q_pred) )
                else:
                    units_str = f" [{mm.quality_unit()}]"
                    print( "{met_name}={Q:0.4f}{units}".format(met_name=mm.short_name(), Q=Q_pred, units=units_str) )
                if not res_fh is None:
                    res_fh.write( f", {Q_pred}" )


                if args.features and not stats is None:
                    if mm == 'pu-psnr':
                        logging.warning( f'Skipping features as it is not supported by {mm}' )
                        break
                    dest_name = os.path.join(out_dir, base + "_fmap.json")
                    logging.info("Writing feature map '" + dest_name + "' ...")
                    mm.write_features_to_json(stats, dest_name)

                if do_heatmap and not stats is None:
                    if stats["heatmap"].shape[2]>1: # if it is a video
                        dest_name = os.path.join(out_dir, base + "_heatmap.mp4")
                        logging.info("Writing heat map '" + dest_name + "' ...")
                        np2vid(torch.squeeze(stats["heatmap"].permute((2,3,4,1,0)), dim=4).cpu().numpy(), dest_name, vs.get_frames_per_second(), args.verbose)
                    else:
                        dest_name = os.path.join(out_dir, base + "_heatmap.png")
                        logging.info("Writing heat map '" + dest_name + "' ...")
                        np2img(torch.squeeze(stats["heatmap"].permute((2,3,4,1,0)), dim=4).cpu().numpy(), dest_name)

                if args.distogram != -1 and not stats is None:
                    dest_name = os.path.join(out_dir, base + "_distogram.png")                    
                    logging.info("Writing distogram '" + dest_name + "' ...")
                    jod_max = args.distogram
                    try: 
                        mm.export_distogram( stats, dest_name, jod_max=jod_max )
                    except NotImplementedError as e:
                        logging.warning( f'Metric {mm.short_name()} cannot generate distograms' )
                    
                del stats

        if not res_fh is None:
            res_fh.write( "\n" )

    if not res_fh is None:
        res_fh.close()

    #     del test_vid
    #     torch.cuda.empty_cache()

def main():
    args = parse_args()

    if args.interactive:
        #print( "Running in an interactive mode" )
        while True:
            line = sys.stdin.readline()
            if not line:
                break

            #print( shlex.split(line) )
            args = parse_args(shlex.split(line))
            run_on_args(args)
    else:
        run_on_args(args)


if __name__ == '__main__':
    main()



import abc

from pycvvdp.video_source import *

# A base class for the video quality metrtics

class vq_metric:

    '''
    test_cont and reference_cont can be either numpy arrays or PyTorch tensors with images or video frames. 
        Depending on the display model (display_photometry), the pixel values should be either display encoded, or absolute linear.
        The two supported datatypes are float16 and uint8.
    dim_order - a string with the order of dimensions of test_cont and reference_cont. The individual characters denote
        B - batch
        C - color channel
        F - frame
        H - height
        W - width
        Examples: "HW" - gray-scale image (column-major pixel order); "HWC" - color image; "FCHW" - color video
        The default order is "BCFHW". The processing can be a bit faster if data is provided in that order. 
    frame_padding - the metric requires at least 250ms of video for temporal processing. Because no previous frames exist in the
        first 250ms of video, the metric must pad those first frames. This options specifies the type of padding to use:
          'replicate' - replicate the first frame
          'circular'  - tile the video in the front, so that the last frame is used for frame 0.
          'pingpong'  - the video frames are mirrored so that frames -1, -2, ... correspond to frames 0, 1, ...
    '''
    def predict(self, test_cont, reference_cont, dim_order="BCFHW", frames_per_second=0, frame_padding="replicate"):

        test_vs = video_source_array( test_cont, reference_cont, frames_per_second, dim_order=dim_order, display_photometry=self.display_photometry )

        return self.predict_video_source(test_vs, frame_padding=frame_padding)

    '''
    The same as `predict` but takes as input fvvdp_video_source_* object instead of Numpy/Pytorch arrays.
    '''
    @abc.abstractmethod
    def predict_video_source(self, vid_source, frame_padding="replicate"):
        pass

    @abc.abstractmethod
    def short_name(self):
        pass

    @abc.abstractmethod
    def quality_unit(self):
        pass

    def get_info_string(self):
        return None

    def set_display_model(self, display_name="standard_4k", display_photometry=None, display_geometry=None, config_paths=[]):
        if display_photometry is None:
            self.display_photometry = vvdp_display_photometry.load(display_name, config_paths)
            self.display_name = display_name
        else:
            self.display_photometry = display_photometry
            self.display_name = "unspecified"

    '''
    Set the base name and path for any extra debuging info or outputs that a metric may produce. 
    '''
    def set_base_fname( self, base_fname ):
        self.base_fname = base_fname

    '''
    Some metrics can be switched between training and inference modes.
    '''
    def train(self, do_training=True):
        pass

    def short_name(self):
        # Class name but '-' instead of '_'
        return self.__class__.__name__.replace('_', '-')


vq_metric_dict = dict()

def register_metric( metric_class ):
    vq_metric_dict[metric_class.__name__] = metric_class    


from abc import abstractmethod
import torch
import os
import numpy as np 
import logging
from torch.functional import Tensor
import pycvvdp.utils as utils
from pycvvdp.display_model import vvdp_display_photometry, vvdp_display_geometry, vvdp_display_photo_eotf

#from pycvvdp.colorspace import ColorTransform

"""
video_source_* objects are used to supply test/reference frames to ColorVideoVDP. 
Those could be coming from memory or files. The subclasses of this abstract class implement
reading the frames and converting them to the approprtate format. 
"""
class video_source:
   
    # Return (height, width, frames) touple with the resolution and
    # the length of the video clip.
    @abstractmethod
    def get_video_size(self):
        pass

    # Return the frame rate of the video
    @abstractmethod
    def get_frames_per_second(self) -> float:
        pass
    
    # Get a test video frame in the selected colorspace. See display_model.py->linear_2_target_colorspace
    # for the list of available color spaces. 
    # You can also pass:
    # 'display_encoded_01', 'display_encoded_100nit' or 'display_encoded_dmax' for the method to return 
    #  display-encoded image (e.g. sRGB) with the values between 0 and 1. If the input source contains linear
    #  values (e.g. an HDR image) they will be PU-encoded:
    # 'display_encoded_01' when the input is 0.005 to 10000, the PU-encoded values are between 0-1
    # 'display_encoded_100nit' when the input is 100, the PU-encoded value is 1, the values can be >1
    # 'display_encoded_dmax' when the input is equal display peak luminance, the PU-encoded value is 1, the values can be >1
    @abstractmethod
    def get_test_frame( self, frame, device, colorspace ) -> Tensor:
        pass

    @abstractmethod
    def get_reference_frame( self, frame, device, colorspace ) -> Tensor:
        pass

    # Check whether pixel values are valid, display warning if it is not the case
    def check_if_valid( self, frame, target_colorspace ):

        if not hasattr( self, "warning_shown" ):
            self.warning_shown = False

        if not self.warning_shown and torch.isnan(frame).flatten().any():
            self.warning_shown = True
            logging.warning( 'Image contains one or more NaN values' )

        if not self.warning_shown and torch.isinf(frame).flatten().any():
            self.warning_shown = True
            logging.warning( 'Image contains one or more Inf values' )

        if not hasattr( self, "first_frame" ):
            self.first_frame = True

        if self.first_frame and not target_colorspace.startswith('display_encoded'):
            self.first_frame = False
            f_mean = torch.mean(frame[:,0,:,:,:])
            f_max = torch.max(frame[:,0,:,:,:])
            f_min = torch.min(frame[:,0,:,:,:])
            logging.debug( f"Content mean={f_mean}, max={f_max}, min={f_min}" )

            if not self.warning_shown and f_mean <= 1:
                logging.warning( 'The mean color value is less than 1 - the image may not be scaled in absolute photometric units!' )

    def get_frame_count(self):
        return self.get_video_size()[2]

    def get_batch_size(self):
        return 1  # Not every video source supports batches.


"""
video_source_filter is used to customize an existing video_source class. All calls are passed via redirection.
"""
class video_source_filter(video_source):

    def __init__(self, vs: video_source):
        self.vs = vs

    # Return (height, width, frames) touple with the resolution and
    # the length of the video clip.
    def get_video_size(self):
        return self.vs.get_video_size()

    # Return the frame rate of the video
    def get_frames_per_second(self) -> float:
        return self.vs.get_frames_per_second()
    
    # Get a test video frame in the selected colorspace. See display_model.py->linear_2_target_colorspace
    # for the list of available color spaces. 
    # You can also pass:
    # 'display_encoded_01', 'display_encoded_100nit' or 'display_encoded_dmax' for the method to return 
    #  display-encoded image (e.g. sRGB) with the values between 0 and 1. If the input source contains linear
    #  values (e.g. an HDR image) they will be PU-encoded:
    # 'display_encoded_01' when the input is 0.005 to 10000, the PU-encoded values are between 0-1
    # 'display_encoded_100nit' when the input is 100, the PU-encoded value is 1, the values can be >1
    # 'display_encoded_dmax' when the input is equal display peak luminance, the PU-encoded value is 1, the values can be >1
    def get_test_frame( self, frame, device, colorspace ) -> Tensor:
        return self.vs.get_test_frame(frame, device, color_space)

    def get_reference_frame( self, frame, device, colorspace ) -> Tensor:
        return self.vs.get_reference_frame(frame, device, color_space)




"""
Function for changing the order of dimensions, for example, from "WHC" (width, height, color) to "BCHW" (batch, color, height, width)
If a dimension is missing in in_dims, it will be added as a singleton dimension
"""
def reshuffle_dims( T: Tensor, in_dims: str, out_dims: str ) -> Tensor:
    in_dims = in_dims.upper()
    out_dims = out_dims.upper()    

    assert len(in_dims) == T.dim(), "The in_dims string must have as many characters as there are dimensions in T"

    # Find intersection of two strings    
    inter_dims = ""
    for kk in range(len(out_dims)):
        if in_dims.find(out_dims[kk]) != -1:
            inter_dims += out_dims[kk]

    # First, squeeze out the dimensions that are missing in the output
    sq_dims = []
    new_in_dims = ""
    for kk in range(len(in_dims)):
        if inter_dims.find(in_dims[kk]) == -1: # The dimension is missing in the output
            sq_dims.append(kk)
            assert T.shape[kk] == 1, "Only the dimensions of size 1 can be skipped in the output"
        else:
            new_in_dims += in_dims[kk]
    in_dims = new_in_dims
    # For the compatibility with PyTorch pre 2.0, squeeze dims one by one
    sq_dims.sort(reverse=True)
    for kk in sq_dims:
        T = T.squeeze(dim=kk)

    # First, permute into the right order
    perm = [0] * len(inter_dims)
    for kk in range(len(inter_dims)):
        ind = in_dims.find(inter_dims[kk])
        assert ind != -1, 'Dimension "{}" missing in the target dimensions: "{}"'.format(in_dims[kk],out_dims)
        perm[kk] = ind                    
    T_p = T.permute(perm)

    # Add missing dimensions
    out_sh = [1] * len(out_dims)
    for kk in range(len(out_dims)):        
        ind = inter_dims.find(out_dims[kk])
        if ind != -1:
            out_sh[kk] = T_p.shape[ind]

    return T_p.reshape( out_sh )


def numpy2torch_frame(np_array, frame, device, dim_order="HWC" ):

    if isinstance( np_array, np.ndarray ):
        if np_array.dtype == np.uint16:
            # Torch does not natively support uint16. A workaround is to pack uint16 values into int16.
            # This will be efficiently transferred and unpacked on the GPU.
            # logging.info('Test has datatype uint16, packing into int16')
            np_array = np_array.astype(np.int16)
        torch_array = torch.tensor(np_array)
    else:
        torch_array = np_array # If it is already a tensor

    from_array = reshuffle_dims( torch_array, in_dims=dim_order, out_dims="BCFHW" )

    if from_array.dtype is torch.float32:
        frame = from_array[:,:,frame:(frame+1),:,:].to(device)
    elif from_array.dtype is torch.float16:
        frame = from_array[:,:,frame:(frame+1),:,:].to(device=device, dtype=torch.float32)
    elif from_array.dtype is torch.int16:
        # Use int16 to losslessly pack uint16 values
        # Unpack from int16 by bit masking as described in this thread:
        # https://stackoverflow.com/a/20766900
        # logging.info('Found int16 datatype, unpack into uint16')
        max_value = 2**16 - 1
        # Cast to int32 to store values >= 2**15
        frame_int32 = from_array[:,:,frame:(frame+1),:,:].to(device).to(torch.int32)
        frame_uint16 = frame_int32 & max_value
        # Finally convert to float in the range [0,1]
        frame = frame_uint16.to(torch.float32) / max_value
    elif from_array.dtype is torch.uint8:
        frame = from_array[:,:,frame:(frame+1),:,:].to(device).to(torch.float32)/255
    else:
        raise RuntimeError( f"Only uint8, uint16 and float32 is currently supported. {from_array.dtype} encountered." )
    return frame


"""
This video_source uses a photometric display model to convert input content (e.g. sRGB) to luminance maps. 
"""
class video_source_dm( video_source ):

    def __init__( self,  display_photometry='sdr_4k_30', config_paths=[] ):

#        self.color_trans = ColorTransform(color_space_name)

        if isinstance( display_photometry, str ):
            self.dm_photometry = vvdp_display_photometry.load(display_photometry, config_paths) 
        elif isinstance( display_photometry, vvdp_display_photometry ):
            self.dm_photometry = display_photometry
        else:
            raise RuntimeError( "display_model must be a string or fvvdp_display_photometry subclass" )

    def apply_dm_and_color_transform(self, frame, target_colorspace):

        I = self.dm_photometry.source_2_target_colorspace(frame, target_colorspace)

        self.check_if_valid(I, target_colorspace)
        return I



"""
This video source supplies frames from either Pytorch tensors and Numpy arrays. It also applies a photometric display model.

A batch of videos should be stored as a tensor or numpy array. Ideally, the tensor should have the dimensions BCFHW (batch, color, frame, height, width).If tensor is stored in another formay, you can pass the order of dimsions as "dim_order" parameter. If any dimension is missing, it will
be added as a singleton dimension. 

This class is for display-encoded (gamma-encoded) content that will be processed by a display model to produce linear  absolute luminance emitted from a display.
"""
class video_source_array( video_source_dm ):
               
    # test_video, reference video - tensor with test and reference video frames. See the class description above for the explanation of dimensions of those tensors.
    # fps - frames per second. Must be 0 for images
    # dim_order - a string with the order of the dimensions. 'BCFHW' is the default.
    # display_model - object that implements fvvdp_display_photometry
    #   class
    # color_space_name - name of the color space (see
    #   fvvdp_data/color_spaces.json)
    def __init__( self, test_video, reference_video, fps, dim_order='BCFHW', display_photometry='sdr_4k_30', config_paths=[], ):

        super().__init__(display_photometry=display_photometry, config_paths=config_paths)        

        if test_video.shape != reference_video.shape:
            raise RuntimeError( 'Test and reference image/video tensors must be exactly the same shape' )
        
        if len(dim_order) != len(test_video.shape):
            raise RuntimeError( 'Input tensor much have exactly as many dimensions as there are characters in the "dims" parameter' )

        # Convert numpy arrays to tensors. Note that we do not upload to device or change dtype at this point (to save GPU memory)
        if isinstance( test_video, np.ndarray ):
            if test_video.dtype == np.uint16:
                # Torch does not natively support uint16. A workaround is to pack uint16 values into int16.
                # This will be efficiently transferred and unpacked on the GPU.
                # logging.info('Test has datatype uint16, packing into int16')
                test_video = test_video.astype(np.int16)
            test_video = torch.tensor(test_video)
        if isinstance( reference_video, np.ndarray ):
            if reference_video.dtype == np.uint16:
                # Torch does not natively support uint16. A workaround is to pack uint16 values into int16.
                # This will be efficiently transferred and unpacked on the GPU.
                # logging.info('Reference has datatype uint16, packing into int16')
                reference_video = reference_video.astype(np.int16)
            reference_video = torch.tensor(reference_video)

        # Change the order of dimension to match BFCHW - batch, frame, color, height, width
        test_video = reshuffle_dims( test_video, in_dims=dim_order, out_dims="BCFHW" )
        reference_video = reshuffle_dims( reference_video, in_dims=dim_order, out_dims="BCFHW" )

        B, C, F, H, W = test_video.shape
        
        if fps==0 and F>1:
            raise RuntimeError( 'When passing video sequences, you must set ''frames_per_second'' parameter' )

        # if C == 4:
        #     logging.warning('Input media has 4 color channels, ignoring the alpha channel to run cvvdp.')
        #     test_video = test_video[:,:3]
        #     reference_video = reference_video[:,:3]
        if C not in (1, 3):
            raise RuntimeError( 'The content must have either 1 or 3 color channels.' )

        self.fps = fps
        self.is_video = (fps>0)
        self.is_color = (C==3)
        self.test_video = test_video
        self.reference_video = reference_video

    def get_frames_per_second(self):
        return self.fps
            
    # Return a [height width frames] vector with the resolution and
    # the number of frames in the video clip. [height width 1] is
    # returned for an image. 
    def get_video_size(self):

        sh = self.test_video.shape
        return (sh[3], sh[4], sh[2])
    
    def get_batch_size(self):
        return self.test_video.shape[0]

    # Get a test video frame as a single-precision luminance map
    # scaled in absolute inits of cd/m^2. 'frame' is the frame index,
    # starting from 0. If use_gpu==true, the function should return a
    # gpuArray.

    def get_test_frame( self, frame, device, colorspace ):
        return self._get_frame(self.test_video, frame, device, colorspace )

    def get_reference_frame( self, frame, device, colorspace ):
        return self._get_frame(self.reference_video, frame, device, colorspace )

    def _get_frame( self, from_array, frame, device, colorspace ):        
        # Determine the maximum value of the data type storing the
        # image/video

        if from_array.dtype is torch.float32:
            frame = from_array[:,:,frame:(frame+1),:,:].to(device)
        elif from_array.dtype is torch.float16:
            frame = from_array[:,:,frame:(frame+1),:,:].to(device=device, dtype=torch.float32)
        elif from_array.dtype is torch.int16:
            # Use int16 to losslessly pack uint16 values
            # Unpack from int16 by bit masking as described in this thread:
            # https://stackoverflow.com/a/20766900
            # logging.info('Found int16 datatype, unpack into uint16')
            max_value = 2**16 - 1
            # Cast to int32 to store values >= 2**15
            frame_int32 = from_array[:,:,frame:(frame+1),:,:].to(device).to(torch.int32)
            frame_uint16 = frame_int32 & max_value
            # Finally convert to float in the range [0,1]
            frame = frame_uint16.to(torch.float32) / max_value
        elif from_array.dtype is torch.uint8:
            frame = from_array[:,:,frame:(frame+1),:,:].to(device).to(torch.float32)/255
        else:
            raise RuntimeError( f"Only uint8, uint16 and float32 is currently supported. {from_array.dtype} encountered." )

        I = self.apply_dm_and_color_transform(frame, colorspace)
        
        return I


class video_source_packed_array( video_source_dm ):
    def __init__(self, test_video, reference_video, fps, display_photometry='sdr_4k_30', config_paths=[], yuv=True, resize_mode='bilinear'):
        super().__init__(display_photometry, config_paths)

        self.fps = fps
        self.is_video = fps > 0
        self.test_video = test_video
        self.reference_video = reference_video
        self.yuv = yuv
        self.resize_mode = resize_mode

    def get_frames_per_second(self):
        return self.fps

    # Return a [height width frames] vector with the resolution and
    # the number of frames in the video clip. [height width 1] is
    # returned for an image.
    def get_video_size(self):
        n, _, _, _, _, h, w = map(int, self.test_video[:7])
        return h, w, n

    def get_test_frame(self, frame, device):
        return self._get_frame(self.test_video, frame, device)

    def get_reference_frame(self, frame, device):
        return self._get_frame(self.reference_video, frame, device)

    def _get_frame(self, from_array, idx, device):
        n, h, w, bit_depth, chroma_ss, resize_h, resize_w = map(int, from_array[:7])
        if self.yuv:
            y_pixels = h*w
            chroma_ss = str(chroma_ss)
            uv_shape = (h//2, w//2) if chroma_ss == '420' else (h, w)
            uv_pixels = uv_shape[0]*uv_shape[1]
            frame_pixels = y_pixels + 2*uv_pixels

            Y = from_array[7 + idx*frame_pixels                        : 7 + idx*frame_pixels + y_pixels]
            u = from_array[7 + idx*frame_pixels + y_pixels             : 7 + idx*frame_pixels + y_pixels + uv_pixels]
            v = from_array[7 + idx*frame_pixels + y_pixels + uv_pixels : 7 + idx*frame_pixels + y_pixels + 2*uv_pixels]

            Yuv_float = utils.fixed2float_upscale(Y, u, v, (h, w), uv_shape, bit_depth, chroma_ss, device)
            if self.color_space=='bt2020nc':
                # display-encoded (PQ) BT.2020 RGB image
                ycbcr2rgb = torch.tensor([[1, 0, 1.47460],
                                        [1, -0.16455, -0.57135],
                                        [1, 1.88140, 0]], device=device)
            else:
                # display-encoded (sRGB) BT.709 RGB image
                ycbcr2rgb = torch.tensor([[1, 0, 1.402],
                                        [1, -0.344136, -0.714136],
                                        [1, 1.772, 0]], device=device)
            RGB = Yuv_float @ ycbcr2rgb.transpose(1, 0)
            # Torch interpolate requires (B,C,H,W)
            RGB = RGB.permute(2,0,1).unsqueeze(0).clip(0, 1)
            if (resize_h != h) or (resize_w != w):
                RGB = torch.nn.functional.interpolate(RGB,
                                                      size=(resize_h, resize_w),
                                                      mode=self.resize_mode)

            # fvvdp required dims -> (B,C,N,H,W)
            frame = RGB.unsqueeze(2)
        else:
            frame_pixels = resize_h*resize_w*3
            if from_array.dtype == np.uint8:
                frame = torch.tensor(from_array[7 + idx*frame_pixels : 7 + (idx+1)*frame_pixels], dtype=torch.uint8)
                frame = frame.to(device).to(torch.float32)
                max_value = 2**8 - 1
            elif from_array.dtype == np.uint16:
                frame = utils.npuint16_to_torchfp32(from_array[7 + idx*frame_pixels : 7 + (idx+1)*frame_pixels], device)
                max_value = 2**16 - 1

            # fvvdp required dims -> (B,C,N,H,W)
            frame = frame.reshape(resize_h, resize_w, 3) / max_value
            frame = frame.permute(2,0,1).reshape(1, -1, 1, resize_h, resize_w)

        L = self.dm_photometry.forward( frame )

        if L.shape[1] == 3:
            # Convert to grayscale
            L = L[:,0:1,:,:,:]*self.color_to_luminance[0] + L[:,1:2,:,:,:]*self.color_to_luminance[1] + L[:,2:3,:,:,:]*self.color_to_luminance[2]

        return L




import scipy.io as spio

# mat load helpers from: https://stackoverflow.com/a/8832212
def loadmat(filename):
    '''
    this function should be called instead of direct spio.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects
    '''
    data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)

def _check_keys(dict):
    '''
    checks if entries in dictionary are mat-objects. If yes
    todict is called to change them to nested dictionaries
    '''
    for key in dict:
        if isinstance(dict[key], spio.matlab.mio5_params.mat_struct):
            dict[key] = _todict(dict[key])
    return dict        

def _todict(matobj):
    '''
    A recursive function which constructs from matobjects nested dictionaries
    '''
    dict = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, spio.matlab.mio5_params.mat_struct):
            dict[strg] = _todict(elem)
        else:
            dict[strg] = elem
    return dict


import torch
from cvvdp_metric import cvvdp

import logging

logging.basicConfig(format='[%(levelname)s] %(message)s', level=logging.DEBUG)

metric = cvvdp(display_name='standard_4k', config_paths=["../metric_configs/cvvdp_overconstancy_transd/cvvdp_parameters.json"])

device = metric.device

sz = (4, 10, 1024, 1024)

with torch.autograd.detect_anomaly():
    for kk in range(100):

        #        "mask_p": { "value": "default", "range":  [2, 3], "clamp": "hard" },
        #        "mask_q": { "value": "default", "range": [1, 2.4], "clamp": "hard" }, 
        #        "xcm_weights": { "value": "default", "range": [-10, 2] }

        metric.mask_p = torch.rand((1), device=device, requires_grad=True)+2
        metric.mask_q = torch.rand((4), device=device, requires_grad=True)*1.4+1
        metric.xcm_weights = torch.rand((4,4), device=device, requires_grad=True)*12-10

        T = torch.randn( sz, device=device )*1
        R = torch.randn( sz, device=device )*1
        S = torch.exp( torch.randn( sz, device=device )*5+3 )

        D = metric.apply_masking_model( T, R, S )

        D_all = D.sum()
        D_all.backward()

        if D.isnan().any() or D.isinf().any():
            print( "Masking produces NaN or Inf values" )
        else:
            print( "No NaN or Inf values found. Huray!" )

    





# Decimated Laplacian Pyramid
from functools import cache
import torch
import torch.nn.functional as Func
import numpy as np 
#import scipy.io as spio
#import os
#import sys
import math
#import torch.autograd.profiler as profiler

def ceildiv(a, b):
    return -(-a // b)

# Decimated Laplacian pyramid
class lpyr_dec():

    def __init__(self, W, H, ppd, device):
        self.device = device
        self.ppd = ppd
        self.min_freq = 0.2
        self.W = W
        self.H = H

        max_levels = int(np.floor(np.log2(min(self.H, self.W))))-1

        bands = np.concatenate([[1.0], np.power(2.0, -np.arange(0.0,14.0)) * 0.3228], 0) * self.ppd/2.0 

        # print(max_levels)
        # print(bands)
        # sys.exit(0)

        invalid_bands = np.array(np.nonzero(bands <= self.min_freq)) # we want to find first non0, length is index+1

        if invalid_bands.shape[-2] == 0:
            max_band = max_levels
        else:
            max_band = invalid_bands[0][0]

        # max_band+1 below converts index into count
        self.height = np.clip(max_band+1, 0, max_levels) # int(np.clip(max(np.ceil(np.log2(ppd)), 1.0)))
        self.band_freqs = np.array([1.0] + [0.3228 * 2.0 **(-f) for f in range(self.height)]) * self.ppd/2.0

        self.pyr_shape = self.height * [None] # shape (W,H) of each level of the pyramid
        self.pyr_ind = self.height * [None]   # index to the elements at each level

        cH = H
        cW = W
        for ll in range(self.height):
            self.pyr_shape[ll] = (cH, cW)
            cH = ceildiv(H,2)
            cW = ceildiv(W,2)

    def get_freqs(self):
        return self.band_freqs

    def get_band_count(self):
        return self.height+1

    def get_band(self, bands, band):
        if band == 0 or band == (len(bands)-1):
            band_mul = 1.0
        else:
            band_mul = 2.0

        return bands[band] * band_mul

    def set_band(self, bands, band, data):
        if band == 0 or band == (len(bands)-1):
            band_mul = 1.0
        else:
            band_mul = 2.0

        bands[band] = data / band_mul

    def get_gband(self, gbands, band):
        return gbands[band]

    # def get_gband_count(self):
    #     return self.height #len(gbands)

    # def clear(self):
    #     for pyramid in self.P:
    #         for level in pyramid:
    #             # print ("deleting " + str(level))
    #             del level

    def decompose(self, image): 
        # assert len(image.shape)==4, "NCHW (C==1) is expected, got " + str(image.shape)
        # assert image.shape[-2] == self.H
        # assert image.shape[-1] == self.W

        # self.image = image

        return self.laplacian_pyramid_dec(image, self.height+1)

    def reconstruct(self, bands):
        img = bands[-1]

        for i in reversed(range(0, len(bands)-1)):
            img = self.gausspyr_expand(img, [bands[i].shape[-2], bands[i].shape[-1]])
            img += bands[i]

        return img

    def laplacian_pyramid_dec(self, image, levels = -1, kernel_a = 0.4):
        gpyr = self.gaussian_pyramid_dec(image, levels, kernel_a)

        height = len(gpyr)
        if height == 0:
            return []

        lpyr = []
        for i in range(height-1):
            layer = gpyr[i] - self.gausspyr_expand(gpyr[i+1], [gpyr[i].shape[-2], gpyr[i].shape[-1]], kernel_a)
            lpyr.append(layer)

        lpyr.append(gpyr[height-1])

        # print("laplacian pyramid summary:")
        # print("self.height = %d" % self.height)
        # print("height      = %d" % height)
        # print("len(lpyr)   = %d" % len(lpyr))
        # print("len(gpyr)   = %d" % len(gpyr))
        # sys.exit(0)

        return lpyr, gpyr

    def interleave_zeros_and_pad(self, x, exp_size, dim):
        new_shape = [*x.shape]
        new_shape[dim] = exp_size[dim]+4
        z = torch.zeros( new_shape, dtype=x.dtype, device=x.device)
        odd_no = (exp_size[dim]%2)
        if dim==-2:
            z[...,2:-2:2,:] = x
            z[...,0,:] = x[...,0,:]
            z[...,-2+odd_no,:] = x[...,-1,:]
        elif dim==-1:
            z[...,:,2:-2:2] = x
            z[...,:,0] = x[...,:,0]
            z[...,:,-2+odd_no] = x[...,:,-1]
        else:
            assert False, "Wrong dimension"

        return z

    def gaussian_pyramid_dec(self, image, levels = -1, kernel_a = 0.4):

        default_levels = int(np.floor(np.log2(min(image.shape[-2], image.shape[-1]))))

        if levels == -1:
            levels = default_levels
        if levels > default_levels:
            raise Exception("Too many levels (%d) requested. Max is %d for %s" % (levels, default_levels, image.shape))

        res = [image]

        for i in range(1, levels):
            res.append(self.gausspyr_reduce(res[i-1], kernel_a))

        return res


    def sympad(self, x, padding, axis):
        if padding == 0:
            return x
        else:
            beg = torch.flip(torch.narrow(x, axis, 0,        padding), [axis])
            end = torch.flip(torch.narrow(x, axis, -padding, padding), [axis])

            return torch.cat((beg, x, end), axis)

    @cache
    def get_kernels( self, im, kernel_a = 0.4 ):

        # ch_dim = len(im.shape)-2
        # if hasattr(self, "K_horiz") and ch_dim==self.K_ch_dim:
        #     return self.K_vert, self.K_horiz

        K = torch.tensor([0.25 - kernel_a/2.0, 0.25, kernel_a, 0.25, 0.25 - kernel_a/2.0], device=im.device, dtype=im.dtype)
        self.K_vert = torch.reshape(K, (1, 1, K.shape[0], 1))
        self.K_horiz = torch.reshape(K, (1, 1, 1, K.shape[0]))
        # self.K_ch_dim = ch_dim
        return self.K_vert, self.K_horiz
        

    def gausspyr_reduce(self, x, kernel_a = 0.4):

        K_vert, K_horiz = self.get_kernels( x, kernel_a )

        H, W = x.shape[-2], x.shape[-1]
        y_a = Func.conv2d(x.view(-1,1,H,W), K_vert, stride=(2,1), padding=(2,0)).view( x.shape[0:-2] + (-1,W) )
        # view(B,C,-1,W)

        # Symmetric padding 
        y_a[...,0,:] += x[...,0,:]*K_vert[...,1,0] + x[...,1,:]*K_vert[...,0,0]
        if (x.shape[-2] % 2)==1: # odd number of rows
            y_a[...,-1,:] += x[...,-1,:]*K_vert[...,3,0] + x[...,-2,:]*K_vert[...,4,0]
        else: # even number of rows
            y_a[...,-1,:] += x[...,-1,:]*K_vert[...,4,0]

        H = y_a.shape[-2]
        y = Func.conv2d(y_a.view(-1,1,H,W), K_horiz, stride=(1,2), padding=(0,2)).view( x.shape[0:-2] + (H,-1) )

        # Symmetric padding 
        y[...,:,0] += y_a[...,:,0]*K_horiz[...,0,1] + y_a[...,:,1]*K_horiz[...,0,0]
        if (x.shape[-2] % 2)==1: # odd number of columns
            y[...,:,-1] += y_a[...,:,-1]*K_horiz[...,0,3] + y_a[...,:,-2]*K_horiz[...,0,4]
        else: # even number of columns
            y[...,:,-1] += y_a[...,:,-1]*K_horiz[...,0,4] 

        return y

    def gausspyr_expand_pad(self, x, padding, axis):
        if padding == 0:
            return x
        else:
            beg = torch.narrow(x, axis, 0,        padding)
            end = torch.narrow(x, axis, -padding, padding)

            return torch.cat((beg, x, end), axis)

    # This function is (a bit) faster
    def gausspyr_expand(self, x, sz = None, kernel_a = 0.4):
        if sz is None:
            sz = [x.shape[-2]*2, x.shape[-1]*2]

        K_vert, K_horiz = self.get_kernels( x, kernel_a )

        y_a = self.interleave_zeros_and_pad(x, dim=-2, exp_size=sz)

        H, W = y_a.shape[-2], y_a.shape[-1]
        y_a = Func.conv2d(y_a.view(-1,1,H,W), K_vert*2).view( x.shape[0:-2] + (-1,W) )

        y   = self.interleave_zeros_and_pad(y_a, dim=-1, exp_size=sz)
        H, W = y.shape[-2], y.shape[-1]

        y   = Func.conv2d(y.view(-1,1,H,W), K_horiz*2).view( x.shape[0:-2] + (H,-1) )

        return y

    def interleave_zeros(self, x, dim):
        z = torch.zeros_like(x, device=self.device)
        if dim==2:
            return torch.cat([x,z],dim=3).view(x.shape[0], x.shape[1], 2*x.shape[2],x.shape[3])
        elif dim==3:
            return torch.cat([x.permute(0,1,3,2),z.permute(0,1,3,2)],dim=3).view(x.shape[0], x.shape[1], 2*x.shape[3],x.shape[2]).permute(0,1,3,2)



# Decimated Laplacian pyramid with a bit better interface - stores all bands within the object
class lpyr_dec_2(lpyr_dec):

    def __init__(self, W, H, ppd, device, keep_gaussian=False):
        self.device = device
        self.ppd = ppd
        self.min_freq = 0.2
        self.W = W
        self.H = H
        self.keep_gaussian=keep_gaussian

        max_levels = int(np.floor(np.log2(min(self.H, self.W))))-1

        bands = np.concatenate([[1.0], np.power(2.0, -np.arange(0.0,14.0)) * 0.3228], 0) * self.ppd/2.0 

        # print(max_levels)
        # print(bands)
        # sys.exit(0)

        invalid_bands = np.array(np.nonzero(bands <= self.min_freq)) # we want to find first non0, length is index+1

        if invalid_bands.shape[-2] == 0:
            max_band = max_levels
        else:
            max_band = invalid_bands[0][0]

        # max_band+1 below converts index into count
        self.height = np.clip(max_band+1, 0, max_levels) # int(np.clip(max(np.ceil(np.log2(ppd)), 1.0)))
        self.band_freqs = np.array([1.0] + [0.3228 * 2.0 **(-f) for f in range(self.height)]) * self.ppd/2.0

        self.pyr_shape = self.height * [None] # shape (W,H) of each level of the pyramid
        self.pyr_ind = self.height * [None]   # index to the elements at each level

        cH = H
        cW = W
        for ll in range(self.height):
            self.pyr_shape[ll] = (cH, cW)
            cH = ceildiv(H,2)
            cW = ceildiv(W,2)

        self.lbands = [None] * (self.height+1) # Laplacian pyramid bands
        if self.keep_gaussian:
            self.gbands = [None] * (self.height+1) # Gaussian pyramid bands

    def get_freqs(self):
        return self.band_freqs

    def get_band_count(self):
        return self.height+1

    def get_lband(self, band):
        if band == 0 or band == (len(self.lbands)-1):
            band_mul = 1.0
        else:
            band_mul = 2.0

        return self.lbands[band] * band_mul

    def set_lband(self, band, data):
        if band == 0 or band == (len(self.lbands)-1):
            band_mul = 1.0
        else:
            band_mul = 2.0

        self.lbands[band] = data / band_mul

    def get_gband(self, band):
        return self.gbands[band]

    # def clear(self):
    #     for pyramid in self.P:
    #         for level in pyramid:
    #             # print ("deleting " + str(level))
    #             del level

    def decompose(self, image): 
        return self.laplacian_pyramid_dec(image, self.height+1)

    def reconstruct(self):
        img = self.lbands[-1]

        for i in reversed(range(0, len(self.lbands)-1)):
            img = self.gausspyr_expand(img, [self.lbands[i].shape[-2], self.lbands[i].shape[-1]])
            img += self.lbands[i]

        return img

    def laplacian_pyramid_dec(self, image, levels = -1, kernel_a = 0.4):
        gpyr = self.gaussian_pyramid_dec(image, levels, kernel_a)

        height = len(gpyr)
        if height == 0:
            return

        lpyr = []
        for i in range(height-1):
            layer = gpyr[i] - self.gausspyr_expand(gpyr[i+1], [gpyr[i].shape[-2], gpyr[i].shape[-1]], kernel_a)
            lpyr.append(layer)

        lpyr.append(gpyr[height-1])
        self.lbands = lpyr

        if self.keep_gaussian:
            self.gbands = gpyr        

        return lpyr, gpyr

# This pyramid computes and stores contrast during decomposition, improving performance and reducing memory consumption
class weber_contrast_pyr(lpyr_dec):

    def __init__(self, W, H, ppd, device, contrast):
        super().__init__(W, H, ppd, device)
        self.contrast = contrast

    def decompose(self, image):
        levels = self.height+1
        kernel_a = 0.4
        gpyr = self.gaussian_pyramid_dec(image, levels, kernel_a)

        height = len(gpyr)
        if height == 0:
            return []

        lpyr = []
        L_bkg_pyr = []
        for i in range(height):
            is_baseband = (i==(height-1))

            if is_baseband:
                layer = gpyr[i]
                if self.contrast.endswith('ref'):
                    L_bkg = torch.clamp(gpyr[i][...,1:2,:,:,:], min=0.01)
                else:
                    # The sustained channels use the mean over the image as the background. Otherwise, they would be divided by itself and the contrast would be 1.
                    L_bkg = torch.mean(torch.clamp(gpyr[i][...,0:2,:,:,:], min=0.01), dim=[-1, -2], keepdim=True)
            else:
                glayer_ex = self.gausspyr_expand(gpyr[i+1], [gpyr[i].shape[-2], gpyr[i].shape[-1]], kernel_a)
                layer = gpyr[i] - glayer_ex 

                # Order: test-sustained-Y, ref-sustained-Y, test-rg, ref-rg, test-yv, ref-yv, test-transient-Y, ref-transient-Y
                # L_bkg is set to ref-sustained 
                if self.contrast == 'weber_g1_ref':
                    L_bkg = torch.clamp(glayer_ex[...,1:2,:,:,:], min=0.01)
                elif self.contrast == 'weber_g1':
                    L_bkg = torch.clamp(glayer_ex[...,0:2,:,:,:], min=0.01)
                elif self.contrast == 'weber_g0_ref':
                    L_bkg = torch.clamp(gpyr[i][...,1:2,:,:,:], min=0.01)
                else:
                    raise RuntimeError( f"Contrast {self.contrast} not supported")

            if L_bkg.shape[-4]==2: # If L_bkg NOT identical for the test and reference images
                contrast = torch.empty_like(layer)
                contrast[...,0::2,:,:,:] = torch.clamp(torch.div(layer[...,0::2,:,:,:], L_bkg[...,0:1,:,:,:]), max=1000.0)    
                contrast[...,1::2,:,:,:] = torch.clamp(torch.div(layer[...,1::2,:,:,:], L_bkg[...,1:2,:,:,:]), max=1000.0)    
            else:
                contrast = torch.clamp(torch.div(layer, L_bkg), max=1000.0)

            lpyr.append(contrast)
            L_bkg_pyr.append(torch.log10(L_bkg))

        # L_bkg_bb = gpyr[height-1][...,0:2,:,:,:]
        # lpyr.append(gpyr[height-1]) # Base band
        # L_bkg_pyr.append(L_bkg_bb) # Base band

        return lpyr, L_bkg_pyr


# This pyramid computes and stores contrast during decomposition, improving performance and reducing memory consumption
class log_contrast_pyr(lpyr_dec):

    def __init__(self, W, H, ppd, device, contrast):
        super().__init__(W, H, ppd, device)
        self.contrast = contrast

        # Assuming D65, there is a linear mapping from log10(L)+log10(M) to log10-luminance
        lms_d65 = [0.7347, 0.3163, 0.0208]
        self.a = 0.5
        self.b = math.log10(lms_d65[0]) - math.log10(lms_d65[1]) + math.log10(lms_d65[0]+lms_d65[1])

    def decompose(self, image):
        levels = self.height+1
        kernel_a = 0.4
        gpyr = self.gaussian_pyramid_dec(image, levels, kernel_a)

        height = len(gpyr)
        if height == 0:
            return []

        lpyr = []
        L_bkg_pyr = []
        for i in range(height):
            is_baseband = (i==(height-1))

            if is_baseband:
                contrast = gpyr[i]
                L_bkg = self.a * (gpyr[i][...,0:2,:,:,:] - self.b)
            else:
                glayer_ex = self.gausspyr_expand(gpyr[i+1], [gpyr[i].shape[-2], gpyr[i].shape[-1]], kernel_a)
                contrast = gpyr[i] - glayer_ex 

                # Order: test-sustained-Y, ref-sustained-Y, test-rg, ref-rg, test-yv, ref-yv, test-transient-Y, ref-transient-Y                
                # Mapping from log10(L) + log10(M) to log10(L+M)
                L_bkg = self.a * (glayer_ex[...,0:2,:,:,:] - self.b)

            lpyr.append(contrast)
            L_bkg_pyr.append(L_bkg)

        
        return lpyr, L_bkg_pyr


    # def gausspyr_expand(self, x, sz = None, kernel_a = 0.4):
    #     if sz is None:
    #         sz = [x.shape[-2]*2, x.shape[-1]*2]

    #     K_vert, K_horiz = self.get_kernels( x, kernel_a )

    #     y_a = self.interleave_zeros(x, dim=2)[...,0:sz[0],:]
    #     y_a = self.gausspyr_expand_pad(y_a, padding=2, axis=-2)
    #     y_a = Func.conv2d(y_a, K_vert*2)

    #     y   = self.interleave_zeros(y_a, dim=3)[...,:,0:sz[1]]
    #     y   = self.gausspyr_expand_pad(  y,   padding=2, axis=-1)
    #     y   = Func.conv2d(y, K_horiz*2)

    #     return y



# if __name__ == '__main__':

#     device = torch.device('cuda:0')

#     torch.set_printoptions(precision=2, sci_mode=False, linewidth=300)

#     image = torch.tensor([
#         [ 1,  2,  3,  4,  5,  6,  7,  8],
#         [11, 12, 13, 14, 15, 16, 17, 18],
#         [21, 22, 23, 24, 25, 26, 27, 28],
#         [31, 32, 33, 34, 35, 36, 37, 38],
#         [41, 42, 43, 44, 45, 46, 47, 48],
#         [51, 52, 53, 54, 55, 56, 57, 58],
#         [61, 62, 63, 64, 65, 66, 67, 68],
#         [71, 72, 73, 74, 75, 76, 77, 78],
#         ], dtype=torch.float32, device=device)

#     image = image.repeat((16, 16))
#     # image = torch.cat((image, image, image), axis = -1)
#     # image = torch.cat((image, image, image), axis = -2)

#     ppd = 50

#     im_tensor = image.view(1, 1, image.shape[-2], image.shape[-1])

#     lp = fvvdp_lpyr_dec_fast(im_tensor.shape[-2], im_tensor.shape[-1], ppd, device)
#     lp_old = fvvdp_lpyr_dec(im_tensor.shape[-2], im_tensor.shape[-1], ppd, device)

#     lpyr, gpyr = lp.decompose( im_tensor )
#     lpyr_2, gpyr_2 = lp_old.decompose( im_tensor )

#     for li in range(lp.get_band_count()):
#         E = Func.mse_loss(lp.get_band(lpyr, li), lp_old.get_band(lpyr_2, li))
#         print( "Level {}, MSE={}".format(li, E))

#     import torch.utils.benchmark as benchmark

#     t0 = benchmark.Timer(
#         stmt='lp.decompose( im_tensor )',
#         setup='',
#         globals={'im_tensor': im_tensor, 'lp': lp})

#     t1 = benchmark.Timer(
#         stmt='lp_old.decompose( im_tensor )',
#         setup='',
#         globals={'im_tensor': im_tensor, 'lp_old': lp_old})

#     print("New pyramid")
#     print(t0.timeit(30))

#     print("Old pyramid")
#     print(t1.timeit(30))

#     # print("----Gaussian----")
#     # for gi in range(lp.get_band_count()):
#     #     print(lp.get_gband(gpyr, gi))

#     # print("----Laplacian----")
#     # for li in range(lp.get_band_count()):
#     #     print(lp.get_band(lpyr, li))




# Visualization of difference maps.

import torch
from pycvvdp.interp import interp1

def luminance_NCHW(x):
    if x.shape[1] == 3: # NC***
        y = (
            x[:,0:1,...] * 0.212656 + 
            x[:,1:2,...] * 0.715158 + 
            x[:,2:3,...] * 0.072186)
    else:
        y = x

    return y

def log_luminance(x):
    y = luminance_NCHW(x)
    clampval = torch.min(y[y>0.0])
    return torch.log(torch.clamp(y, min=clampval))


def vis_tonemap(b, dr):
    t = 3.0
    
    b_min = torch.min(b)
    b_max = torch.max(b)
    
    if b_max-b_min < dr: # No tone-mapping needed
        tmo_img = (b-b_min)/(b_max-b_min+1e-3)*dr + (1-dr)/2
        return tmo_img

    b_scale = torch.linspace( b_min, b_max, 1024, device=b.device)
    b_p = torch.histc( b, 1024, b_min, b_max )
    b_p = b_p / torch.sum(b_p)
    
    sum_b_p = torch.sum(torch.pow(b_p, 1.0/t))

    dy = torch.pow(b_p, 1.0/t) / sum_b_p
    
    v = torch.cumsum(dy, 0)*dr + (1.0-dr)/2.0
    
    tmo_img = interp1(b_scale, v, b)

    return tmo_img

# returns sRGB image/frames
def visualize_diff_map(diff_map, context_image=None, colormap_type="supra-threshold", use_cpu=False):
    diff_map = torch.clamp(diff_map, 0.0, 1.0)
    if use_cpu:
        diff_map = diff_map.cpu()
        context_image = context_image.cpu()

    if context_image is None:
        tmo_img = torch.ones_like(diff_map) * 0.5
    else:
        tmo_img = vis_tonemap(log_luminance(context_image), 0.6)

    if colormap_type == 'threshold':
        # Visualize up to 1 JOD (>=1 JOD will be all red)

        color_map = torch.tensor([
            [0.2, 0.2, 1.0],
            [0.2, 1.0, 1.0],
            [0.2, 1.0, 0.2],
            [1.0, 1.0, 0.2],
            [1.0, 0.2, 0.2],
        ], device=diff_map.device)
        color_map_in = torch.tensor([0.00, 0.25, 0.50, 0.75, 1.00], device=diff_map.device)*0.1

    elif colormap_type == 'supra-threshold':
        # Visualize up to 3 JOD (>=3 JOD will be all yellow)

        color_map = torch.tensor([
            [0.2, 1.0, 1.0],
            [1.0, 1.0, 1.0],
            [1.0, 1.0, 0.2],
        ], device=diff_map.device)
        color_map_in = torch.tensor([0.0, 0.5, 1.0], device=diff_map.device)*0.3

    elif colormap_type == 'monochromatic':
        
        color_map = torch.tensor([
            [1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0],
        ], device=diff_map.device)
        
        color_map_in = torch.tensor([0.0, 1.0], device=diff_map.device)
    else:
        print("Unknown colormap: %s" % colormap_type)

    # cmap = torch.zeros_like(diff_map)
    # if cmap.shape[1] == 1:
    #     cmap = torch.cat([cmap]*3, 1)
    frame_count, h, w = diff_map.shape[-3], diff_map.shape[-2], diff_map.shape[-1]
    cmap = torch.empty( [3, frame_count, h, w], device=diff_map.device, dtype=torch.float16)

    color_map_l = color_map[:,0:1] * 0.212656 + color_map[:,1:2] * 0.715158 + color_map[:,2:3] * 0.072186
    color_map_ch = color_map / (torch.cat([color_map_l] * 3, 1) + 0.0001)

    cmap[0:1,...] = interp1(color_map_in, color_map_ch[:,0], diff_map).type(torch.float16)
    cmap[1:2,...] = interp1(color_map_in, color_map_ch[:,1], diff_map).type(torch.float16)
    cmap[2:3,...] = interp1(color_map_in, color_map_ch[:,2], diff_map).type(torch.float16)

    cmap = (cmap * tmo_img).clip(0.,1.)

    return cmap


#from pycvvdp import colorspace
import torch

from pycvvdp.utils import PU
from pycvvdp.video_source import *
from pycvvdp.vq_metric import *


"""
Plain PSNR-RGB metric. Operates on display-encoded values. If HDR/linear color is encountered, it will be 
PU21-encoded. The display model is used only for images in linear color spaces. Usage is same as 
the ColorVideoVDP metric (see pytorch_examples).
"""
class psnr_rgb(vq_metric):

    def __init__(self, display_name="standard_4k", display_photometry=None, device=None, config_paths=[]):
        # Use GPU if available
        if device is None:
            if torch.cuda.is_available() and torch.cuda.device_count()>0:
                self.device = torch.device('cuda:0')
            else:
                self.device = torch.device('cpu')
        else:
            self.device = device

        self.set_display_model( display_name=display_name, display_photometry=display_photometry, config_paths=config_paths )

    '''
    The same as `predict` but takes as input fvvdp_video_source_* object instead of Numpy/Pytorch arrays.
    '''
    def predict_video_source(self, vid_source, frame_padding="replicate"):

        _, _, N_frames = vid_source.get_video_size()

        batch_sz = vid_source.get_batch_size()
        mse = torch.zeros((batch_sz), device=self.device)
        for ff in range(N_frames):
            # colorspace='display_encoded_100nit' will get us display-encoded image, or if the original source is linear, it will apply PU-encoding.
            # If the input is PQ-encoded, it will return a PQ-encoded values. 
            T = vid_source.get_test_frame(ff, device=self.device, colorspace='display_encoded_100nit')
            R = vid_source.get_reference_frame(ff, device=self.device, colorspace='display_encoded_100nit')
            mse += torch.mean( (T - R)**2, dim=(1,2,3,4) )

        max_I = 1
        psnr = 20*torch.log10( max_I/torch.sqrt(mse/N_frames) ) 
        
        return psnr, None

    def short_name(self):
        return "PSNR-RGB"

    def quality_unit(self):
        return "dB"

register_metric( psnr_rgb )

"""
PU21-PSNR-Y metric. Usage is same as the ColorVideoVDP metric (see pytorch_examples).
"""
class pu_psnr_y(vq_metric):

    def __init__(self, display_name="standard_4k", display_photometry=None, color_space="sRGB", device=None, config_paths=[]):
        # Use GPU if available
        if device is None:
            if torch.cuda.is_available() and torch.cuda.device_count()>0:
                self.device = torch.device('cuda:0')
            else:
                self.device = torch.device('cpu')
        else:
            self.device = device

        self.set_display_model( display_name=display_name, display_photometry=display_photometry, config_paths=config_paths )
        self.color_space = color_space # input content color space

        self.pu = PU()
        self.max_I = self.pu.encode(torch.as_tensor(100)) # 100 nit should correspond to white on an SDR display
        self.metric_colorspace = 'Y' # color space in which the metric operates

    '''
    The same as `predict` but takes as input fvvdp_video_source_* object instead of Numpy/Pytorch arrays.
    '''
    def predict_video_source(self, vid_source, frame_padding="replicate"):

        _, _, N_frames = vid_source.get_video_size()
        batch_sz = vid_source.get_batch_size()

        mse = torch.zeros((batch_sz), device=self.device)
        for ff in range(N_frames):
            T = vid_source.get_test_frame(ff, device=self.device, colorspace=self.metric_colorspace)
            R = vid_source.get_reference_frame(ff, device=self.device, colorspace=self.metric_colorspace)

            # Apply PU
            T_enc = self.pu.encode(T)
            R_enc = self.pu.encode(R)

            mse += torch.mean( (T - R)**2, dim=(1,2,3,4) )
        
        psnr = 20*torch.log10( self.max_I/torch.sqrt(mse/N_frames) ) 

        return psnr, None

    def psnr_fn(self, img1, img2):
        mse = torch.mean( (img1 - img2)**2 )
        return 20*torch.log10( self.pu.peak/torch.sqrt(mse) )

    def short_name(self):
        return "PU21-PSNR-Y"

    def quality_unit(self):
        return "dB"

register_metric( pu_psnr_y )


class pu_psnr_rgb2020(pu_psnr_y):
    def __init__(self, display_name="standard_4k", display_photometry=None, color_space="sRGB", device=None):
        super().__init__(display_name=display_name, display_photometry=display_photometry, color_space=color_space, device=device)
        self.metric_colorspace = 'RGB2020'

    def short_name(self):
        return "PU21-PSNR-RGB2020"

register_metric( pu_psnr_rgb2020 )



import torch

from pycvvdp.utils import PU
from pycvvdp.video_source import *
from pycvvdp.vq_metric import *

from pycvvdp.third_party.ssim import SSIM

def get_luma(img):
    return 0.212656*img[...,0,:,:,:] + 0.715158 * img[...,1,:,:,:] + 0.072186 * img[...,2,:,:,:]

"""
Plain SSIM metric, computed on the luma channel. Operates on display-encoded values. If HDR/linear color is encountered, it will be 
PU21-encoded. The display model is used only for images in linear color spaces. Usage is same as 
the ColorVideoVDP metric (see pytorch_examples).
"""
class ssim_metric(vq_metric):

    def __init__(self, display_name="standard_4k", display_photometry=None, color_space="sRGB", device=None):
        # Use GPU if available
        if device is None:
            if torch.cuda.is_available() and torch.cuda.device_count()>0:
                self.device = torch.device('cuda:0')
            else:
                self.device = torch.device('cpu')
        else:
            self.device = device

        self.set_display_model( display_name=display_name, display_photometry=display_photometry )
        self.color_space = color_space # input content color space
        self.ssim = SSIM(channel=1, data_range=1.)


    '''
    The same as `predict` but takes as input fvvdp_video_source_* object instead of Numpy/Pytorch arrays.
    '''
    def predict_video_source(self, vid_source, frame_padding="replicate"):

        _, _, N_frames = vid_source.get_video_size()
        
        ssim_index = 0 
        n = 0
        for ff in range(N_frames):
            # colorspace='display_encoded_100nit' will get us display-encoded image, or if the original source is linear, it will apply PU-encoding.
            # If the input is PQ-encoded, it will return a PQ-encoded values (). 
            T = get_luma(vid_source.get_test_frame(ff, device=self.device, colorspace='display_encoded_100nit'))
            R = get_luma(vid_source.get_reference_frame(ff, device=self.device, colorspace='display_encoded_100nit'))

            ssim_index += self.ssim.forward(T, R)
            n += 1
        
        return ssim_index/n, None

    def short_name(self):
        return "SSIM"

    def quality_unit(self):
        return ""




import os
import torch
import numpy as np
import json
import torch.nn.functional as Func
import math
from functools import cache

from pycvvdp.interp import interp1, interp1q
#from PIL import Image

from pycvvdp.third_party.loadmat import loadmat

def torch_gpu_mem_info():
    t = torch.cuda.get_device_properties(0).total_memory
    c = torch.cuda.memory_cached(0)
    a = torch.cuda.memory_allocated(0)
    f = c-a  # free inside cache
    print("GPU mem used: %d M (cache %d M)" % (a/(1024*1024), c/(1024*1024)))

def json2dict(file):
    data = None
    if os.path.isfile(file):
        with open(file, "r") as json_file:
            data=json.load(json_file)
    else:
        raise RuntimeError( f"Error: Cannot find file {file}" )
    return data

def linear2srgb_torch(lin):
    lin = torch.clamp(lin, 0.0, 1.0)
    srgb = torch.where(lin > 0.0031308, (1.055 * (lin ** (1/2.4))) - 0.055, 12.92 * lin)
    return srgb

def srgb2linear_torch(srgb):
    srgb = torch.clamp(srgb, 0.0, 1.0)
    lin = torch.where(srgb > 0.04045, ((srgb + 0.055) / 1.055)**2.4, srgb/12.92)
    return lin

def img2np(img):
    return np.array(img, dtype="float32") * 1.0/255.0

# def np2img(nparr):
#     return Image.fromarray(np.clip(nparr * 255.0, 0.0, 255.0).astype('uint8'))

def l2rgb(x):
    return np.concatenate([x,x,x], -1)

def stack_horizontal(nparr):
    return np.concatenate([nparr[i] for i in range(len(nparr))], axis=-2)

def stack_vertical(nparr):
    return np.concatenate([nparr[i] for i in range(len(nparr))], axis=-3)


def load_mat_dict(filepath, data_label, device):
    # datapath = "D:\\work\\st_fov_metric"
    # filepath = os.path.join(datapath, rel_path)

    if not os.path.isfile(filepath):
        return None
    else:
        v = loadmat(filepath)
        if data_label in v:
            return v[data_label]
        else:
            print("Cannot find key %s, valid keys are %s" % (data_label, v.keys()))
            return None

def load_mat_tensor(filepath, data_label, device):
    # datapath = "D:\\work\\st_fov_metric"
    # filepath = os.path.join(datapath, rel_path)

    if not os.path.isfile(filepath):
        return None
    else:
        v = loadmat(filepath)
        if data_label in v:
            return torch.tensor(v[data_label], device=device)
        else:
            print("Cannot find key %s, valid keys are %s" % (data_label, v.keys()))
            return None


# args are indexed at 0
def fovdots_load_ref(content_id, device, data_res="full"):
    if data_res != "full":
        print("Note: Using data resolution %s" % (data_res))
    hwd_tensor = load_mat_tensor("D:\\work\\st_fov_metric\\data_vid_%s\\content_%d_ref.mat" % (data_res, content_id+1), "I_vid", device)
    dhw_tensor = hwd_tensor.permute(2,0,1)
    ncdhw_tensor = torch.unsqueeze(torch.unsqueeze(dhw_tensor, 0), 0)
    return ncdhw_tensor

# args are indexed at 0
def fovdots_load_condition(content_id, condition_id, device, data_res="full"):
    if data_res != "full":
        print("Note: Using data resolution %s" % (data_res))
    hwd_tensor = load_mat_tensor("D:\\work\\st_fov_metric\\data_vid_%s\\content_%d_condition_%d.mat" % (data_res, content_id+1, condition_id+1), "I_vid", device)
    dhw_tensor = hwd_tensor.permute(2,0,1)
    ncdhw_tensor = torch.unsqueeze(torch.unsqueeze(dhw_tensor, 0), 0)
    return ncdhw_tensor


class ImGaussFilt():
    def __init__(self, sigma, device):
        self.filter_size = 2 * int(np.ceil(2.0 * sigma)) + 1
        self.half_filter_size = (self.filter_size - 1)//2

        self.K = torch.zeros((1, 1, self.filter_size, self.filter_size), device=device)

        for ii in range(self.filter_size):
            for jj in range(self.filter_size):
                distsqr = float(ii - self.half_filter_size) ** 2 + float(jj - self.half_filter_size) ** 2
                self.K[0,0,jj,ii] = np.exp(-distsqr / (2.0 * sigma * sigma))

        self.K = self.K/self.K.sum()

    def run(self, img):
        
        if len(img.shape) == 2: img_4d = img.reshape((1,1,img.shape[0],img.shape[1]))
        else:                   img_4d = img

        pad = (
            self.half_filter_size,
            self.half_filter_size,
            self.half_filter_size,
            self.half_filter_size,)

        img_4d = Func.pad(img_4d, pad, mode='reflect')
        return Func.conv2d(img_4d, self.K)[0,0]


class config_files:
    # fvvdp_config_dir = None
    
    # @classmethod
    # def set_config_dir( cls, path ):
    #     cls.fvvdp_config_dir = path

    @classmethod
    def find(cls, fname, config_paths:list):

        if not isinstance( config_paths, list ):
            raise RuntimeError( "config_paths must be a list" )

        bname, ext = os.path.splitext(fname)
        # First check if the matching file name is in the config paths
        for cp in config_paths:
            if not (os.path.isfile(cp) or os.path.isdir(cp)):
                raise RuntimeError( f"config_path '{cp}' does not exist" )

            if os.path.isfile(cp) and os.path.basename(cp).startswith(bname):
                return cp

        # Then, check all directories
        for cp in config_paths:
            if os.path.isdir(cp):
                path = os.path.join( cp, fname )
                if os.path.isfile(path):
                    return path

        # Then, check CVVDP_PATH
        ev_config_dir = os.getenv("CVVDP_PATH")
        if not ev_config_dir is None:
            path = os.path.join( ev_config_dir, fname )
            if os.path.isfile(path):
                return path

        # Finally, check the default config dir
        path = os.path.join(os.path.dirname(__file__), "vvdp_data", fname)
        if os.path.isfile(path):
            return path

        raise RuntimeError( f"The configuration file {fname} not found" )


class PU():
    '''
    Transform absolute linear luminance values to/from the perceptually uniform space.
    This class is intended for adopting image quality metrics to HDR content.
    This is based on the new spatio-chromatic CSF from:
      Wuerger, S., Ashraf, M., Kim, M., Martinovic, J., Perez-Ortiz, M., & Mantiuk, R. K. (2020).
      Spatio-chromatic contrast sensitivity under mesopic and photopic light levels.
      Journal of Vision, 20(4), 23. https://doi.org/10.1167/jov.20.4.23
    The implementation should work for both numpy arrays and torch tensors
    '''
    def __init__(self, L_min=0.005, L_max=10000, type='banding_glare'):
        self.L_min = L_min
        self.L_max = L_max

        if type == 'banding':
            self.p = [1.070275272, 0.4088273932, 0.153224308, 0.2520326168, 1.063512885, 1.14115047, 521.4527484]
        elif type == 'banding_glare':
            self.p = [0.353487901, 0.3734658629, 8.277049286e-05, 0.9062562627, 0.09150303166, 0.9099517204, 596.3148142]
        elif type == 'peaks':
            self.p = [1.043882782, 0.6459495343, 0.3194584211, 0.374025247, 1.114783422, 1.095360363, 384.9217577]
        elif type == 'peaks_glare':
            self.p = [816.885024, 1479.463946, 0.001253215609, 0.9329636822, 0.06746643971, 1.573435413, 419.6006374]
        else:
            raise ValueError(f'Unknown type: {type}')

        self.peak = self.p[6]*(((self.p[0] + self.p[1]*L_max**self.p[3])/(1 + self.p[2]*L_max**self.p[3]))**self.p[4] - self.p[5])

    def _encode_direct(self, Y):
        '''
        Convert from linear (optical) values Y to encoded (electronic) values V
        '''
        # epsilon = 1e-5
        # if (Y < (self.L_min - epsilon)).any() or (Y > (self.L_max + epsilon)).any():
        #     print( 'Values passed to encode are outside the valid range' )
        Y = Y.clip(self.L_min, self.L_max)
        Y_p = Y**self.p[3]
        V = self.p[6]*(((self.p[0] + self.p[1]*Y_p)/(1 + self.p[2]*Y_p))**self.p[4] - self.p[5])
        return V

    @cache
    def _get_encode_lut(self, device):
        Y_lut = torch.linspace(math.log10(self.L_min), math.log10(self.L_max), 2048, device=device)
        V_lut = self._encode_direct(10**Y_lut)
        return (Y_lut, V_lut)

    def encode(self, Y):
        '''
        Convert from linear (optical) values Y to encoded (electronic) values V
        '''
        if True or Y.numel()<100: # Interpolation seems slower than directly computing the values
            V = self._encode_direct(Y)
        else:
            (Y_lut, V_lut) = self._get_encode_lut(Y.device)
            V = interp1q( Y_lut, V_lut, torch.log10(Y.clamp(self.L_min, self.L_max)) )
        return V

    def decode(self, V):
        '''
        Convert from encoded (electronic) values V into linear (optical) values Y
        '''
        V_p = ((V/self.p[6] + self.p[5]).clip(min=0))**(1/self.p[4])
        Y = ((V_p - self.p[0]).clip(min=0)/(self.p[1] - self.p[2]*V_p))**(1/self.p[3])
        return Y



import numpy as np
import ffmpeg

class VideoWriter:

    """
    fps - frames per second
    hdr_mode - encode HDR moview
    codec - 'vp9' (works with Chrome) or 'h265' (everything else) - currently used only in the HDR mode
    """
    def __init__(self, fname, fps=24, hdr_mode=False, codec='h265', verbose=False):
        self.fname = fname
        self.fps = fps
        self.verbose = verbose
        self.hdr_mode = hdr_mode
        self.process = None
        #self.bit_depth = 10
        self.codec = codec


    """
    Write a frame stored as numpy WxHxC matric to a video file. The frame must be in the right, display-encoded colour space:
    BT.709 + sRGB nonlinearity for SDR
    BT.2020 + PQ for HDR
    """
    def write_frame_rgb(self, rgb):
        H, W, C = rgb.shape        
        if C == 1:
            rgb = np.concatenate([rgb]*3, -1)

        if self.process is None:
            if self.hdr_mode:
                if self.codec == 'h265':
                    self.process = (ffmpeg
                            .input('pipe:', format='rawvideo', pix_fmt='rgb48le', s='{}x{}'.format(W, H), r=self.fps, colorspace="bt2020nc", color_primaries="bt2020", color_trc="smpte2084")
                            .output(self.fname, pix_fmt='yuv420p10le', crf=12, vcodec='libx265', **{'x265-params': 'hdr-opt=1:repeat-headers=1:colorprim=bt2020:transfer=smpte2084:colormatrix=bt2020nc:master-display=G(0,0)B(0,0)R(0,0)WP(0,0)L(0,0):max-cll=0,0'} )
                            .overwrite_output()
                            .global_args( '-loglevel', 'info' if self.verbose else 'warning')
                            .global_args( '-hide_banner')
                            .global_args( '-preset', 'fast' )
                            .run_async(pipe_stdin=True, quiet=not self.verbose)
                            )
                elif self.codec == 'vp9':
                    self.process = (ffmpeg
                            .input('pipe:', format='rawvideo', pix_fmt='rgb48le', s='{}x{}'.format(W, H), r=self.fps, colorspace="bt2020nc", color_primaries="bt2020", color_trc="smpte2084")
                            .output(self.fname, pix_fmt='yuv420p10le', crf=10, vcodec='libvpx-vp9', **{'color_primaries': '9', 'color_trc': '16', 'colorspace': '9', 'color_range': '1', 'profile:v': '2' , 'preset': 'fast', 'b:v': '0' } )
                            .overwrite_output()
                            .global_args( '-hide_banner')
                            .global_args( '-preset', 'fast' )
                            .global_args( '-loglevel', 'info' if self.verbose else 'quiet')
                            .run_async(pipe_stdin=True)
                            )
                else:
                    raise RuntimeError( 'Unknown codec' )
            else:
                self.process = (ffmpeg
                        .input('pipe:', format='rawvideo', pix_fmt='rgb24', s='{}x{}'.format(W, H), r=self.fps)
                        .output(self.fname, pix_fmt='yuv420p', **{ "c:v": "mpeg4", "qscale:v": "3" } ) #format='mp4', crf=10                        
                        .overwrite_output()
                        .global_args( '-hide_banner')
                        .global_args( '-loglevel', 'info' if self.verbose else 'quiet')
                        .run_async(pipe_stdin=True)
                )
        

        #ffmpeg.compile(self.process)

        if self.hdr_mode:
            self.process.stdin.write(
                (rgb * (2**16-1)).astype(np.uint16).tobytes()
                )            
        else:
            if rgb.dtype == np.uint8:
                self.process.stdin.write( rgb.tobytes() )
            else:
                self.process.stdin.write(
                    (rgb * 255.0)
                    .astype(np.uint8)
                    .tobytes()
                )

    # Delete or close if program was interrupted
    def __del__(self):
        self.close()        

    def close(self):
        if not self.process is None:
            self.process.stdin.close()
            self.process.wait()        
            self.process = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()




from video_source import *
import re

import logging

def decode_video_props( fname ):
    vprops = dict()
    vprops["width"]=1920
    vprops["height"]=1080

    vprops["fps"] = 24
    vprops["bit_depth"] = 8
    vprops["color_space"] = '2020'
    vprops["chroma_ss"] = '420'

    bname = os.path.splitext(os.path.basename(fname))[0]
    fp = bname.split("_")

    res_match = re.compile( r'(\d+)x(\d+)p?' )

    for field in fp:

        if res_match.match( field ):
            res = field.split( "x")
            if len(res) != 2:
                raise ValueError("Cannot decode the resolution")
            vprops["width"]=int(res[0])
            vprops["height"]=int(res[1])
            continue

        if field.endswith("fps"):
            vprops["fps"] = float(field[:-3])

        if field=="444" or field=="420":
            vprops["chroma_ss"]=field

        if field=="10" or field=="10b":
            vprops["bit_depth"]=10

        if field=="8" or field=="8b":
            vprops["bit_depth"]=8

        if field=="2020" or field=="709":
            vprops["color_space"]=field

        if field=="bt709":
            vprops["color_space"]="709"

        if field=="ct2020" or field=="pq2020":
            vprops["color_space"]="2020"

    return vprops

# Create a filename which encodes the yuv header. It can be parsed with decode_video_props and pfstools.
def create_yuv_fname( basename, vprops ):
    width = vprops["width"]
    height = vprops["height"]
    bit_depth = vprops["bit_depth"]
    color_space = vprops["color_space"]
    chroma_ss = vprops["chroma_ss"]
    fps = vprops["fps"]
    fps = round(fps,3) if round(fps)!=fps else int(fps)  #do not use decimals if not needed
    yuv_name = f"{basename}_{width}x{height}_{bit_depth}b_{chroma_ss}_{color_space}_{fps}fps.yuv"
    return yuv_name


class YUVReader:

    def __init__(self, file_name):        
        self.file_name = file_name

        if not os.path.isfile(file_name):
            raise FileNotFoundError( "File {} not found".format(file_name) )

        vprops = decode_video_props(file_name)

        self.width = vprops["width"]
        self.height = vprops["height"]
        self.avg_fps = vprops["fps"]
        self.color_space = vprops["color_space"]
        self.chroma_ss = vprops["chroma_ss"]

        self.bit_depth = vprops["bit_depth"]
        self.frame_bytes = int(self.width*self.height)
        self.y_pixels = int(self.frame_bytes)
        self.y_shape = (vprops["height"], vprops["width"])

        if vprops["chroma_ss"]=="444":
            self.frame_bytes *= 3
            self.uv_pixels = self.y_pixels
            self.uv_shape = self.y_shape
        else: # Chroma sub-sampling
            self.frame_bytes = self.frame_bytes*3/2
            self.uv_pixels = int(self.y_pixels/4)
            self.uv_shape = (int(self.y_shape[0]/2), int(self.y_shape[1]/2))

        self.frame_pixels = self.frame_bytes
        if vprops["bit_depth"]>8:
            self.frame_bytes *= 2
            self.dtype = np.uint16
        else:
            self.dtype = np.uint8

        self.frames = os.stat(file_name).st_size / self.frame_bytes
#        if math.ceil(self.frame_count)!=self.frame_count:
#            raise RuntimeError( ".yuv file does not seem to contain an integer number of frames" )

        self.frames = int(self.frames)

        self.mm = None

    def get_frame_count(self):
        return int(self.frames)
    
    def get_frame_yuv( self, frame_index ):

        if frame_index<0 or frame_index>=self.frames:
            raise RuntimeError( "The frame index is outside the range of available frames")

        if self.mm is None: # Mem-map as needed
            self.mm = np.memmap( self.file_name, self.dtype, mode="r")

        offset = int(frame_index*self.frame_pixels)
        Y = self.mm[offset:offset+self.y_pixels]
        u = self.mm[offset+self.y_pixels:offset+self.y_pixels+self.uv_pixels]
        v = self.mm[offset+self.y_pixels+self.uv_pixels:offset+self.y_pixels+2*self.uv_pixels]

        return (np.reshape(Y,self.y_shape,'C'),np.reshape(u,self.uv_shape,'C'),np.reshape(v,self.uv_shape,'C'))

    # Return display-encoded (sRGB) BT.709 RGB image
    def get_frame_rgb_rec709( self, frame_index ):

        (Y,u,v) = self.get_frame_yuv(frame_index)

        YUV = fixed2float( np.concatenate( (Y[:,:,np.newaxis],\
            convert420to444(u)[:,:,np.newaxis],\
            convert420to444(v)[:,:,np.newaxis]), axis=2), self.bit_depth)

        RGB = (np.reshape( YUV, (self.y_pixels, 3), order='F' ) @ ycbcr2rgb_rec709.transpose()).reshape( (*self.y_shape, 3 ), order='F' )

        return RGB

    # Return display-encoded (PQ) BT.2020 RGB image
    def get_frame_rgb_rec2020( self, frame_index ):

        (Y,u,v) = self.get_frame_yuv(frame_index)

        YUV = fixed2float( np.concatenate( (Y[:,:,np.newaxis],\
            convert420to444(u)[:,:,np.newaxis],\
            convert420to444(v)[:,:,np.newaxis]), axis=2), self.bit_depth)

        RGB = (np.reshape( YUV, (self.y_pixels, 3), order='F' ) @ ycbcr2rgb_rec2020.transpose()).reshape( (*self.y_shape, 3 ), order='F' )

        return RGB

    # Return RGB PyTorch tensor
    def get_frame_rgb_tensor( self, frame_index, device ):

        if frame_index<0 or frame_index>=self.frames:
            raise RuntimeError( "The frame index is outside the range of available frames")

        if self.mm is None: # Mem-map as needed
            self.mm = np.memmap( self.file_name, self.dtype, mode="r")

        offset = int(frame_index*self.frame_pixels)
        Y = self.mm[offset:offset+self.y_pixels]
        u = self.mm[offset+self.y_pixels:offset+self.y_pixels+self.uv_pixels]
        v = self.mm[offset+self.y_pixels+self.uv_pixels:offset+self.y_pixels+2*self.uv_pixels]

        Yuv_float = self._fixed2float_upscale(Y, u, v, device)

        if self.color_space=='2020':
            # display-encoded (PQ) BT.2020 RGB image
            ycbcr2rgb = torch.tensor([[1, 0, 1.47460],
                                        [1, -0.16455, -0.57135],
                                        [1, 1.88140, 0]], device=device)
        else:
            # display-encoded (sRGB) BT.709 RGB image
            ycbcr2rgb = torch.tensor([[1, 0, 1.402],
                                    [1, -0.344136, -0.714136],
                                    [1, 1.772, 0]], device=device)

        RGB = Yuv_float @ ycbcr2rgb.transpose(1, 0)
        return RGB.clip(0, 1)


    def _np_to_torchfp32(self, X, device):
        if X.dtype == np.uint8:
            return torch.tensor(X, dtype=torch.uint8).to(device).to(torch.float32)
        elif X.dtype == np.uint16:
            return self._npuint16_to_torchfp32(X, device)

    # Torch does not natively support uint16. A workaround is to pack uint16 values into int16.
    # This will be efficiently transferred and unpacked on the GPU.
    # logging.info('Test has datatype uint16, packing into int16')
    def _npuint16_to_torchfp32(self, np_x_uint16, device):
        max_value = 2**16 - 1
        assert np_x_uint16.dtype == np.uint16
        np_x_int16 = torch.tensor(np_x_uint16.astype(np.int16), dtype=torch.int16)
        torch_x_int32 = np_x_int16.to(device).to(torch.int32)
        torch_x_uint16 = torch_x_int32 & max_value
        torch_x_fp32 = torch_x_uint16.to(torch.float32)
        return torch_x_fp32

    def _fixed2float_upscale(self, Y, u, v, device):
        offset = 16/219
        weight = 1/(2**(self.bit_depth-8)*219)
        Yuv = torch.empty(self.height, self.width, 3, device=device)

        Y = self._np_to_torchfp32(Y, device)
        Yuv[..., 0] = torch.clip(weight*Y - offset, 0, 1).reshape(self.height, self.width)

        offset = 128/224
        weight = 1/(2**(self.bit_depth-8)*224)

        uv = np.stack((u, v))
        uv = self._np_to_torchfp32(uv, device)
        uv = torch.clip(weight*uv - offset, -0.5, 0.5).reshape(1, 2, self.uv_shape[0], self.uv_shape[1])

        if self.chroma_ss=="420":
            # TODO: Replace with a proper filter.
            uv_upscaled = torch.nn.functional.interpolate(uv, scale_factor=2, mode='bilinear')
        else:
            uv_upscaled = uv

        Yuv[...,1:] = uv_upscaled.squeeze().permute(1,2,0)

        return Yuv

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.mm = None


"""
This class is compatible with ffmpeg video readers - supports resizing and can be used with video_source_video_file
"""
class video_reader_yuv(YUVReader):

    def __init__(self, vidfile, frames=-1, resize_fn=None, resize_height=-1, resize_width=-1, verbose=False):
        super().__init__(vidfile)
        self.resize_fn=resize_fn
        self.resize_width = resize_width
        self.resize_height = resize_height        
        self.color_transfer = None
        if frames!=-1:
            self.frames = min(self.frames, frames)
        self.curr_frame = -1

    def get_frame(self):
        self.curr_frame += 1
        return self.curr_frame       

    def unpack(self, frame_index, device):
        RGB = self.get_frame_rgb_tensor(frame_index, device)

        if not self.resize_fn is None and (vid_reader.height != self.resize_height or vid_reader.width != self.resize_width):
            RGB = torch.nn.functional.interpolate(RGB.view(1,RGB.shape[0],RGB.shape[1],RGB.shape[2]),
                                                size=(self.resize_height, self.resize_width),
                                                mode=self.resize_fn).view(RGB_bcfhw.shape[0],self.resize_height,self.resize_width).clip(0.,1.)
        return RGB


class video_source_yuv_file(video_source_dm):

    def __init__( self, test_fname, reference_fname, display_photometry='standard_4k', frames=-1, full_screen_resize=None, resize_resolution=None, retain_aspect_ratio=False, verbose=False ):

        self.reference_vidr = YUVReader(reference_fname)
        self.test_vidr = YUVReader(test_fname)
        self.total_frames = self.test_vidr.frames
        self.frames = self.total_frames if frames==-1 else min(self.total_frames, frames)
        self.offset = 0     # Offset for random access of a shorter subsequence

        self.full_screen_resize = full_screen_resize
        if retain_aspect_ratio:
            h, w = self.test_vidr.height, self.test_vidr.width
            if h / resize_resolution[1] * resize_resolution[0] <= w:
                # retain provided width: resize_resolution[0]
                resize_resolution = (resize_resolution[0], int(resize_resolution[0] / w * h))
            else:
                # retain provided height: resize_resolution[1]
                resize_resolution = (int(resize_resolution[1] / h * w), resize_resolution[1])

        self.resize_resolution = resize_resolution

        # if color_space_name=='auto':
        #     if self.test_vidr.color_space=='2020':
        #         color_space_name="BT.2020"
        #     else:
        #         color_space_name="sRGB"

        super().__init__(display_photometry=display_photometry)        

        for vr in [self.test_vidr, self.reference_vidr]:
            if vr == self.test_vidr:
                logging.debug(f"Test video '{test_fname}':")
            else:
                logging.debug(f"Reference video '{reference_fname}':")
            if full_screen_resize is None:
                rs_str = ""
            else:
                rs_str = f"->[{resize_resolution[0]}x{resize_resolution[1]}]"
            logging.debug(f"  [{vr.width}x{vr.height}]{rs_str}, colorspace: {vr.color_space}, EOTF: {self.dm_photometry.EOTF}, fps: {vr.avg_fps}, frames: {self.frames}" )

        
    # Return (height, width, frames) touple with the resolution and
    # the length of the video clip.
    def get_video_size(self):
        if not self.full_screen_resize is None:
            return [self.resize_resolution[1], self.resize_resolution[0], self.frames]
        else:
            return [self.test_vidr.height, self.test_vidr.width, self.frames]

    # Return the frame rate of the video
    def get_frames_per_second(self) -> int:
        return self.test_vidr.avg_fps
    
    # Get a pair of test and reference video frames as a single-precision luminance map
    # scaled in absolute inits of cd/m^2. 'frame' is the frame index,
    # starting from 0. 
    def get_test_frame( self, frame, device, colorspace="Y" ) -> Tensor:
        L = self._get_frame( self.test_vidr, frame, device, colorspace )
        return L

    def get_reference_frame( self, frame, device, colorspace="Y" ) -> Tensor:
        L = self._get_frame( self.reference_vidr, frame, device, colorspace )
        return L

    def _get_frame( self, vid_reader, frame, device, colorspace="Y" ):
        RGB = vid_reader.get_frame_rgb_tensor(self.offset + frame, device)
        RGB_bcfhw = reshuffle_dims( RGB, in_dims='HWC', out_dims="BCFHW" )

        if not self.full_screen_resize is None and (vid_reader.height != self.resize_resolution[1] or vid_reader.width != self.resize_resolution[0]):
            RGB_bcfhw = torch.nn.functional.interpolate(RGB_bcfhw.view(1,RGB_bcfhw.shape[1],RGB_bcfhw.shape[3],RGB_bcfhw.shape[4]),
                                                size=(self.resize_resolution[1], self.resize_resolution[0]),
                                                mode=self.full_screen_resize).view(1,RGB_bcfhw.shape[1],1,self.resize_resolution[1],self.resize_resolution[0]).clip(0.,1.)

        I = self.apply_dm_and_color_transform(RGB_bcfhw, colorspace)
        return I
    
    def set_offset( self, offset:int ):
        self.offset = offset

    # Depreciated
    def get_total_frames(self):
        return self.total_frames

    def set_num_frames(self, num_frames:int):
        if self.offset + num_frames > self.total_frames:
            logging.error(f'Cannot set num_frames={num_frames} because offset={self.offset} and total_frames={self.total_frames}. '
                          f'Clipping num_frames to {self.total_frames - self.offset}')
            num_frames = self.total_frames - self.offset
        self.frames = num_frames